from collections.abc import Generator

from .....mcp_manager.client import ClientManagerMCP
from ...prompts import system_prompts, user_prompts
from ......config.llm_config import ConfigGPT
from ......config.logger import logger
from ...llm import LLM
from openai import OpenAI, AsyncOpenAI
import json


class GPT(LLM):
    """
    ## GPT
    ### Args:
    - `model`: modelo de gpt que a usar
    - `max_history_len`: largo maximo del historial que se para como input al modelo
    """

    def __init__(
        self,
        model=ConfigGPT.DEFAULT_MODEL_NAME,
        max_history_len: int = 10,
        chat_history: list = [],
    ):
        super().__init__()
        self.client = OpenAI(api_key=ConfigGPT.OPENAI_API_KEY)
        """Cliente secuencial de GPT"""
        self.async_client = AsyncOpenAI(api_key=ConfigGPT.OPENAI_API_KEY)
        """Cliente asincrono de GPT"""
        self.max_len_history: int = max_history_len
        """Maxima cantidad de mensajes previos que se le pasan como input"""
        self.chat_history: list[list[dict[str, str]]] = chat_history
        """Historial del chat asosiado a esta instancia de GPT, en forma lista de listas, por ejemplo 
        ```
        chat_history: str = [
            [
                {"role": "user", "content": "query1"},
                {"role": "assistant", "content": "response1"}
            ],
            [
                {"role": "user", "content": "query2"},
                {"role": "assistant", "content": "response2"}
            ]
        ]
        ```
        """
        self.model: str = model
        self.current_price: float = 0

    def __get_price(self, usage) -> float:
        """
        ### Args
            - `usage`: uso de la api retornado en el completion respuesta del llamado a la api de gpt
        ### Outs:
            - `price`: precio final del llamado a la api.
        """
        try:
            input_tokens: int = usage.prompt_tokens
            output_tokens: int = usage.completion_tokens

            input_price: float = ConfigGPT.MODEL_PRICE[self.model]["input"]
            output_price: float = ConfigGPT.MODEL_PRICE[self.model]["output"]
            price: float = input_tokens * input_price + output_tokens * output_price

            # Aumenta el valod asociado al costo de uso de la API en esta instancia de GPT
            self.current_price += price
            return price
        except Exception as e:
            logger.warning(f"Can't calculate usage price from: {self.model}")
            return 0.0

    def append_chat_history(self):
        """Agrega una historia vacia para crear el espacio"""
        self.chat_history.append([])
        self.chat_history[-1].append({})

    def __call_stream_completion(
        self,
        system_message: str,
        query: str,
        extra_messages: list[dict[str, str]] = [],
    ):
        """ """
        self.chat_history[-1][0] = {"role": "user", "content": query}
        messages: list[dict[str, str]] = [{"role": "system", "content": system_message}]

        for message in extra_messages:
            if message["role"] == "system":
                messages.append(message)
        messages += [
            message
            for messages in self.chat_history[-self.max_len_history : -1]
            for message in messages
        ]
        for message in extra_messages:
            if message["role"] == "user" or message["role"] == "assistant":
                messages.append(message)
        messages.append(self.chat_history[-1][0])

        stream = self.client.chat.completions.create(
            model=self.model,
            messages=messages,
            stream=True,
        )
        return stream

    def __call_completion(
        self,
        system_message: str,
        query: str,
        json_format: bool = False,
        extra_messages: list[dict[str, str]] = [],
    ):
        """ """
        self.chat_history[-1][0] = {"role": "user", "content": query}
        messages: list[dict[str, str]] = [{"role": "system", "content": system_message}]

        for message in extra_messages:
            if message["role"] == "system":
                messages.append(message)
        messages += [
            message
            for messages in self.chat_history[-self.max_len_history : -1]
            for message in messages
        ]
        for message in extra_messages:
            if message["role"] == "user" or message["role"] == "assistant":
                messages.append(message)
        messages.append(self.chat_history[-1][0])

        if json_format:
            completion = self.client.chat.completions.create(
                model=self.model,
                messages=messages,
                response_format={"type": "json_object"},
            )
        else:

            completion = self.client.chat.completions.create(
                model=self.model,
                messages=messages,
            )
        self.__get_price(completion.usage)
        return completion.choices[0].message.content

    def preprocess_query(self, query: str) -> dict:
        system_message: str = system_prompts.preproccess_query(
            services=self.client_manager_mcp.get_services()
        )
        messages: list[dict[str, str]] = [
            {"role": "system", "content": system_message},
            {"role": "user", "content": query},
        ]
        completion = self.client.chat.completions.create(
            model=self.model,
            messages=messages,
            response_format={"type": "json_object"},
        )
        self.__get_price(completion.usage)
        response = completion.choices[0].message.content
        return json.loads(response)

    def select_prompts(self, query: str, extra_messages: list[dict[str, str]]) -> str:
        system_message: str = system_prompts.select_prompts
        prompts = self.client_manager_mcp.get_prompts()
        if prompts is None or len(prompts) == 0:
            return json.dumps({"prompt_services": []})

        extra_messages: list[dict[str, str]] = extra_messages + [
            {
                "role": "user",
                "content": user_prompts.exposed_prompts(prompts),
            }
        ]

        return self.__call_completion(
            system_message=system_message,
            query=query,
            json_format=True,
            extra_messages=extra_messages,
        )

    def select_service(
        self,
        query: str,
        extra_messages: list[dict[str, str]] = [],
    ) -> str:
        system_message: str = system_prompts.select_service
        services = self.client_manager_mcp.get_services()
        if services is None or len(services) == 0:
            return json.dumps({"service": "", "args": {}})

        query: str = user_prompts.query_and_services(query=query, services=services)
        return self.__call_completion(
            system_message=system_message,
            query=query,
            json_format=True,
            extra_messages=extra_messages,
        )

    def simple_query(
        self,
        query: str,
        use_services_contex: bool = False,
        extra_messages: list[dict[str, str]] = [],
    ) -> Generator[str, None]:
        system_message: str = system_prompts.chat_asistant(
            self.client_manager_mcp.get_services() if use_services_contex else None
        ) + system_prompts.language_prompt(self.current_language)
        # response = self.__call_completion(system_message=system_message, query=query)

        stream = self.__call_stream_completion(
            system_message=system_message,
            query=query,
            extra_messages=extra_messages,
        )

        response: str = ""
        for chunk in stream:
            content = chunk.choices[0].delta.content
            response += content if content is not None else ""
            yield content

        # Se agrega la respuesta a la historia
        self.chat_history[-1].append({"role": "assistant", "content": response})
        return response

    def final_response(
        self,
        query: str,
        data: str | dict,
        extra_messages: list[dict[str, str]] = [],
    ) -> Generator[str, None]:
        system_message: (
            str
        ) = system_prompts.chat_asistant() + system_prompts.language_prompt(
            self.current_language
        )
        user_message = user_prompts.query_and_data(query=query, data=data)

        stream = self.__call_stream_completion(
            system_message=system_message,
            query=user_message,
            extra_messages=extra_messages,
        )

        response: str = ""
        for chunk in stream:
            content = chunk.choices[0].delta.content
            response += content if content is not None else ""
            yield content

        # Se agrega la respuesta a la historia
        self.chat_history[-1].append({"role": "assistant", "content": response})
        return response

    async def close(self) -> None:
        """
        Closes and cleans up all GPT resources including OpenAI clients and chat history.

        This method performs the following cleanup operations:
        - Closes the async OpenAI client if present
        - Closes the sync OpenAI client if present
        - Clears the chat history to free memory
        - Nullifies client references to prevent memory leaks

        Called by the parent Fastchat cleanup process to ensure proper resource management.
        """
        try:
            # Close async OpenAI client
            if hasattr(self, "async_client") and self.async_client is not None:
                await self.async_client.close()
                self.async_client = None

            # Close sync OpenAI client
            if hasattr(self, "client") and self.client is not None:
                # Sync client doesn't have async close method, so we just nullify
                self.client = None

            # Clear chat history to free memory
            if hasattr(self, "chat_history"):
                self.chat_history.clear()

            # Clear client manager reference
            if hasattr(self, "client_manager_mcp"):
                self.client_manager_mcp = None

        except Exception as e:
            logger.error(f"Error during GPT cleanup: {e}")
            raise



from dotenv import load_dotenv

load_dotenv()


from fastapi import FastAPI
from fastapi.responses import RedirectResponse
from fastauth import Fastauth, TokenRouter, FastauthSettings
from .settings import FastappSettings
from .routes.chat import router as chat_router
from ..config import ConfigLLM, AuthApiConfig
from ..utils.clear_console import clear_console


class FastApp:
    def __init__(
        self,
        extra_reponse_system_prompts: list[str] = [],
        extra_selection_system_prompts: list[str] = [],
        len_context: int = ConfigLLM.DEFAULT_HISTORY_LEN,
        token_router: TokenRouter = TokenRouter(),
        auth_settings: FastauthSettings | dict | None = AuthApiConfig().auth_settings,
    ):
        FastappSettings.update(
            extra_reponse_system_prompts,
            extra_selection_system_prompts,
            len_context,
        )

        
        auth_settings.access_token_paths += ["/chat/user"]
        auth_settings.master_token_paths += ["/chat/admin"]

        self.fastauth = Fastauth(settings=auth_settings)
        self.token_router = token_router

    @property
    def app(self) -> FastAPI:
        app: FastAPI = FastAPI()

        @app.get("/")
        async def root():
            return RedirectResponse(url="/docs")

        @app.get("/healt")
        async def healt():
            return {"status": "healtly"}

        app.include_router(chat_router)
        self.fastauth.set_auth(app, [self.token_router.route])
        return app

    def run(self, host: str, port: int):
        clear_console()
        pass



import os
import json
from mcp_oauth import OAuthClient
from ...config.logger import logger


class Servers:
    """
    Manages server configurations and OAuth clients for the fastchat-mcp application.
    This class loads server credentials and configuration from a JSON file, initializes
    OAuth clients for each server, and provides access to server information.
    Attributes:
        json_config (dict): The loaded configuration from the JSON file.
        app_name (str): The name of the application.
        mcp_servers (dict[str, dict] | None): Dictionary containing server configurations.
    Methods:
        __init__(config_file_path: str = "fastchat.config.json", app_name: str = "fastchat-mcp"):
            Initializes the Servers instance, loads configuration, and sets up servers.
        __load_config_file(config_file_path: str):
            Loads and parses the configuration JSON file.
        __load_servers():
            Loads server credentials from the configuration and initializes OAuth clients.
        __create_oauth_client(server: dict[str, dict]) -> None:
            Creates an OAuth client for the specified server using provided credentials.
    """

    def __init__(
        self,
        config_file_path: str = "fastchat.config.json",
        app_name: str = "fastchat-mcp",
        aditional_servers: dict = {},
    ):
        """
        Initializes the Servers instance.
        Loads the configuration from the specified JSON file, sets the application name,
        and initializes the server credentials and OAuth clients.
        Args:
            config_file_path (str): Path to the configuration JSON file.
            app_name (str): Name of the application.
        """
        # self.config_file_path: str = config_file_path
        self.json_config: dict = self.__load_config_file(
            config_file_path=config_file_path
        )
        self.app_name = (
            self.json_config["app_name"] if "app_name" in self.json_config else app_name
        )
        self.mcp_servers: dict[str, dict] | None = {}
        self.__load_servers(aditional_servers)

    def __load_config_file(self, config_file_path: str):
        """
        Loads and parses the configuration JSON file.
        Args:
            config_file_path (str): Path to the configuration JSON file.
        Returns:
            dict: Parsed configuration data.
        """

        path = f"{os.getcwd()}{os.path.sep}{config_file_path}"
        if not os.path.exists(path=path):
            logger.warning("config file does not exist")
            return {}

        with open(path, "r") as file:
            json_config: dict = json.loads(file.read())
        return json_config

    def __load_servers(self, aditional_servers: dict = {}):
        """
        Loads server credentials from the configuration.
        Initializes OAuth clients for each server defined in the configuration file.
        """

        self.mcp_servers = self.json_config["mcp_servers"] | aditional_servers

        for server in self.mcp_servers.values():
            self.__create_oauth_client(server=server)

    def __create_oauth_client(self, server: dict[str, dict]) -> None:
        """
        Creates an OAuth client for the specified server.
        Uses the credentials provided in the configuration to generate an OAuth client
        and attaches it to the server's configuration.
        Args:
            server (dict[str, dict]): Server configuration dictionary.
        """
        oauth_client: OAuthClient | None = None

        server_url: str | None = (
            server["httpstream-url"] if "httpstream-url" in server else None
        )

        if "auth" in server.keys() and (
            "required" in server["auth"].keys() and server["auth"]["required"]
        ):
            auth: dict = server["auth"]
            post_body = auth["post_body"]
            oauth_client = OAuthClient(
                client_name=self.app_name,
                mcp_server_url=server_url,
                body=post_body,
            )

        server["oauth_client"] = oauth_client



import yaml
import importlib.resources

# from pathlib import Path
# yaml_file = Path(__file__).parent / "texts" / "system_prompts.yaml"
# with open(yaml_file) as f:
    # system_prompts: dict = yaml.safe_load(f)

with importlib.resources.open_text('fastchat.app.services.llm.prompts.texts', 'system_prompts.yaml') as f:
    system_prompts: dict = yaml.safe_load(f)


def language_prompt(language: str) -> str:
    return f"\nAlways respond in the language: {language}"


def chat_asistant(services: list | None = None) -> str:
    return system_prompts.get("chat_assistant_prompt", "") + (
        ""
        if services is None
        else f"\n You have access to the following aviable services:\n{services}"
    )

    # return """
    #     You are an advanced assistant specialized in connecting to and operating with MCP services. Your main function is to always respond based on the services and data available in the MCP context, prioritizing this information over any general model knowledge. Data provided by the user must be interpreted and processed as structured and objective system information, never as subjective messages or personal interpretations from the user.

    #     When you identify events, states, or errors in the received data, respond directly and technically, explaining the cause and possible solutions according to the contextual information, without making assumptions about the user's intent. For example, if you detect an error, respond: "An error occurred in [detail] due to [probable cause] or [alternative]," instead of attributing it to what it 'seems' according to the user.

    #     You must always:
    #     - Prioritize contextual and MCP services information in your responses.
    #     - Use the general knowledge of the LLM only as a backup when the MCP context does not provide sufficient information.
    #     - Maintain a professional and clear tone, adapting the technical level to the user.
    #     - Provide structured, precise, and useful responses, focused on problem resolution or explanation of system states.
    #     """ + (
    #     ""
    #     if services is None
    #     else f"\n You have access to the following aviable services:\n{services}"
    # )


select_service: str = """
You are an expert in data comprehension with access to various services. You receive a user query and a JSON containing information about all available services in the format {key1: data1, key2: data2, ...}, where each value contains the necessary information about the service (name, description, and arguments).

Your task is:
1. Identify if the user's query requests to use any service.
   - If so, select the key of the most useful service for the context and extract the necessary arguments to execute it from the provided data.
   - If there is no useful service or the query only asks for information about the services (without requesting their use), select an empty service ("") and the arguments should be {}; {"service":"", "args":{}}.

2. ALWAYS return a JSON in the following format:
{"service":"service_key", "args":{...}}

Correctly select and extract both the service and the arguments from the input JSON.
"""

select_prompts: str = """
You are an expert in data comprehension with access to various prompt_services. You receive a user query and a JSON containing information about all available prompt_services in the format {key1: data1, key2: data2, ...}, where each value contains the necessary information about the prompt_service (name, description, and arguments).

Your task is:
1. Identify if the user's query requests to use any prompt_service.
   - If so, select the keys of the mosts useful prompt_services for the context and extract the necessary arguments to execute each from the provided data.
   - If there is no useful prompt_service select an empty prompt_services: {prompt_services: []}.

2. ALWAYS return a JSON in the following format:
{
    "prompt_services":
        [
            {
                "prompt_service":"prompt_service_1_key", 
                "args":{...}
            },
            {
                "prompt_service":"prompt_service_2_key", 
                "args":{...}
            },
            ...
        ]
}

Correctly select and extract both the prompt_services and the arguments from the input JSON.
"""


def preproccess_query(services: list) -> str:
    return system_prompts["task_query_decomposer"].replace("{services}", str(services))

    # return (
    #     """You are an expert in task comprehension and ordering of the same. Your mission is, given a user's query, to separate it into several independent queries if necessary. If the query doesn't need to be separated into more than one task then you must return a list of size 1 with exactly the same user's query. It's important that you separate them in the correct order of execution according to their dependencies on each other. The condition for separating queries is given by a list of available services, if it's necessary to use more than one service then you must separate the query. Also you must extract the language used  in the query, it can be any language.\n"""
    #     + "\nBe sure to use a language consistent with the query you are analyzing. If the query is expressed in natural language, you must keep the subqueries in natural language; for example: you cannot split a natural-language query into technically worded subqueries."  # NEW
    #     + """For example, if the query is: "Extract the information of the user with id 222 from database 1 and add this user to database 2" and you have a service to consult the database and another to add to the database; then you must separate the query into two subqueries: {"querys":["Extract the information of the user with id 222 from database 1", "Add this user's information to database 2"],"language":"language used in the query"}.\n"""
    #     + 'You must return the response in a JSON format with the structure: `{"querys":[list of each of the resulting queries],"language":"language used in the query"}`.\n'
    #     + f"The list of available services is the following:\n {services}"
    # ).replace("'", '"')



import os
from ..app.chat.features.llm_provider import LLMProvider


class ConfigLLM:
    DEFAULT_PROVIDER: LLMProvider = LLMProvider.OPENAI
    DEFAULT_HISTORY_LEN: int = 20


class ConfigGPT:
    MODEL_PRICE = {
        "gpt-3.5-turbo": {
            "input": 0.5 / 1000000,
            "output": 1.5 / 1000000,
        },
        "gpt-4o-mini": {
            "input": 0.15 / 1000000,
            "output": 0.60 / 1000000,
        },
        "gpt-5-mini": {
            "input": 0.25 / 1000000,
            "output": 2.00 / 1000000,
        },
        "gpt-5-nano": {
            "input": 0.05 / 1000000,
            "output": 0.40 / 1000000,
        },
    }
    """Price by one tokens from each model"""

    # DEFAULT_MODEL_NAME = "gpt-4o-mini"
    DEFAULT_MODEL_NAME = "gpt-5-nano"
    """GPT model that will be used"""

    OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY")
    """OpenAI Api Key"""



import os
from mcp import StdioServerParameters
import shutil


def get_stdio_session_params(server) -> StdioServerParameters:
    config: dict = server["config"]

    server_params: StdioServerParameters = StdioServerParameters(
        command=(
            shutil.which("npx") if config["command"] == "npx" else config["command"]
        ),
        args=config["args"],
        env={**os.environ, **config["env"]} if config.get("env") else None,
    )
    return server_params



from mcp.client.streamable_http import streamablehttp_client
from mcp.client.auth import OAuthClientProvider
from mcp import ClientSession
import asyncio

from mcp_oauth import OAuthClient

# CLEANUP IMPLEMENTATION CHANGE:
# Global dictionary to store active sessions for proper cleanup
# This allows us to track and close HTTP/WebSocket connections later
_active_sessions: dict[str, tuple] = {}


def get_session_data(
    http: str,
    oauth_client: OAuthClient,
    headers: dict[str, str] | None = None,
) -> dict:
    try:
        loop = asyncio.get_event_loop()
        future = asyncio.run_coroutine_threadsafe(
            async_get_session(
                http=http,
                oauth_client=oauth_client,
                headers=headers,
            ),
            loop,
        )
        result = future.result()  # Espera el resultado
        return result
    except RuntimeError:
        return asyncio.run(
            async_get_session(
                http=http,
                oauth_client=oauth_client,
                headers=headers,
            )
        )


async def async_get_session(
    http: str,
    oauth_client: OAuthClient,
    headers: dict[str, str] | None = None,
):
    # CLEANUP IMPLEMENTATION CHANGE:
    # Generate session key for tracking active connections
    session_key = f"{http}_{id(oauth_client)}_{id(headers)}"
    
    # Connect to a streamable HTTP server
    oauth: OAuthClientProvider | None = (
        oauth_client.oauth if oauth_client is not None else None
    )
    
    # CLEANUP IMPLEMENTATION CHANGE:
    # Instead of using context managers that close immediately,
    # manually manage the connection lifecycle for later cleanup
    connection = streamablehttp_client(url=http, auth=oauth, headers=headers)
    read_stream, write_stream, client = await connection.__aenter__()
    
    # Create session
    session = ClientSession(read_stream, write_stream)
    await session.__aenter__()
    
    # CLEANUP IMPLEMENTATION CHANGE:
    # Store active session references for cleanup tracking
    # This allows close_session() to properly terminate connections
    _active_sessions[session_key] = (connection, session, client)
    
    try:
        # Initialize the connection
        await session.initialize()
        # Generate data
        tools = await session.list_tools()
        resources = await session.list_resource_templates()
        prompts = await session.list_prompts()
        data: dict = {
            "tools": tools.tools,
            "resources": resources.resourceTemplates,
            "prompts": prompts.prompts,
            "_session_key": session_key,  # Include key for cleanup tracking
        }
        return data
    except Exception as e:
        # If error, cleanup immediately to prevent resource leaks
        await close_session_by_key(session_key)
        raise e


async def close_session(server: dict) -> None:
    """
    Close HTTP stream session for the given server.
    
    NEW FUNCTION ADDED FOR CLEANUP IMPLEMENTATION:
    This function closes HTTP/WebSocket connections for a specific MCP server.
    It reconstructs the session key from server configuration and calls
    close_session_by_key to perform the actual cleanup.
    
    Args:
        server (dict): Server configuration containing connection details
    """
    # Reconstruct session key from server config
    http = server.get("httpstream-url", "")
    oauth_client = server.get("oauth_client")
    headers = server.get("headers")
    
    session_key = f"{http}_{id(oauth_client)}_{id(headers)}"
    
    await close_session_by_key(session_key)


async def close_session_by_key(session_key: str) -> None:
    """
    Close session by its key.
    
    NEW FUNCTION ADDED FOR CLEANUP IMPLEMENTATION:
    This function performs the actual cleanup of HTTP/WebSocket connections.
    It properly closes the session and connection using their __aexit__ methods
    and removes the session from the active sessions dictionary.
    
    Args:
        session_key (str): The session key to close
    """
    if session_key in _active_sessions:
        connection, session, client = _active_sessions[session_key]
        
        try:
            # Close session properly using context manager protocol
            await session.__aexit__(None, None, None)
        except Exception:
            pass  # Ignore cleanup errors
        
        try:
            # Close connection properly using context manager protocol
            await connection.__aexit__(None, None, None)
        except Exception:
            pass  # Ignore cleanup errors
        
        # Remove from active sessions tracking
        del _active_sessions[session_key]


async def close_all_sessions() -> None:
    """
    Close all active HTTP sessions.
    
    NEW FUNCTION ADDED FOR CLEANUP IMPLEMENTATION:
    This function closes all currently active HTTP/WebSocket sessions.
    Useful for complete cleanup when shutting down the application.
    """
    session_keys = list(_active_sessions.keys())
    for session_key in session_keys:
        await close_session_by_key(session_key)



import asyncio

from ..app.chat.chat import Fastchat
from ..utils.clear_console import clear_console
from ..config.llm_config import ConfigGPT, ConfigLLM
from ..config.logger import logger
import os


class TerminalChat:
    def __init__(
        self,
        model: str = ConfigGPT.DEFAULT_MODEL_NAME,
        extra_reponse_system_prompts: list[str] = [],
        extra_selection_system_prompts: list[str] = [],
        aditional_servers: dict = {},
        len_context: int = ConfigLLM.DEFAULT_HISTORY_LEN,
        logger_lv: str = "ERROR",
    ):
        """
        Initialize the class with model configuration and system prompts.
        Args:
            model (str): The name of the model to use. Defaults to ConfigGPT.DEFAULT_MODEL_NAME.
            extra_reponse_system_prompts (list[str]): Additional system prompts for responses. Defaults to an empty list.
            extra_selection_system_prompts (list[str]): Additional system prompts for MCP services selection. Defaults to an empty list.
            len_context (int): The maximum length of the context history. Defaults to ConfigLLM.DEFAULT_HISTORY_LEN.
        """

        self.model: str = model
        self.extra_reponse_system_prompts: list[str] = extra_reponse_system_prompts
        self.extra_selection_system_prompts: list[str] = extra_selection_system_prompts
        self.aditional_servers: dict = aditional_servers
        self.len_context: int = len_context
        logger.setLevel(logger_lv)

    def open(self):
        return asyncio.run(self.__open())

    async def __open(self):
        """
        ### open_local_chat
        - Launches an interactive local chat session in the console.
        - The conversation is also recorded in Markdown format. When the user exits (by typing 'exit' or pressing Enter),
        the chat history is saved to a Markdown file in the 'chats' directory, using the first query as the filename.
        """

        clear_console()
        chat: Fastchat = Fastchat(
            model=self.model,
            extra_reponse_system_prompts=self.extra_reponse_system_prompts,
            extra_selection_system_prompts=self.extra_selection_system_prompts,
            aditional_servers=self.aditional_servers,
            len_context=self.len_context,
        )
        await chat.initialize()
        print("\n")

        md: str = ""
        filename: str | None = None
        sep = os.path.sep

        while True:
            query = input("> ")
            if query == "exit" or query == "":
                if md != "":
                    chats_path = f"{os.getcwd()}{sep}chats{sep}"
                    os.makedirs(os.path.dirname(chats_path), exist_ok=True)
                    with open(f"{chats_path}{filename}.md", "w") as file:
                        file.write(md)
                break

            if filename is None:
                filename = f"{query[:15]}..."

            md += f"# {query}"
            index = 1
            async for step in chat(query):
                md += str(step)

                if step.type == "response":
                    if step.json["first_chunk"]:
                        print(f"<< {step.response}", end="")
                    else:
                        print(f"{step.response}", end="")

                if step.type == "query":
                    print(f">> {step.query}")
                    index = 1
                if step.type == "data":
                    print(f'   {str(step).replace("**", "").replace("- ", "• ")}')
                if step.type == "step":
                    print(f"  {index}. {step.step}")
                    index += 1
            md += "\n"



from mcp_oauth import OAuthClient
from mcp.client.auth import OAuthClientProvider
from mcp.client.streamable_http import streamablehttp_client
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client
import asyncio
import shutil
import os


from ..service import Service
from .....utils.stdio_session_params import get_stdio_session_params


class Prompt(Service):
    """Represents a prompt that can be called with arguments."""

    def __init__(self, data, server):
        super().__init__(data, server)
        self.args: list[dict] = [
            {"name": arg.name, "type": "string"} for arg in data.arguments
        ]

    async def __call__(self, args: dict[str, any]):
        return await self.get(args)

    async def get(self, args: dict[str, any]):
        args = {key: str(args[key]) for key in args.keys()}

        if self.protocol == "httpstream":
            return await self.__httpstream_get(
                self.http,
                self.name,
                args,
                self.oauth_client,
                self.headers,
            )

        if self.protocol == "stdio":
            return await self.__stdio_get(
                promptname=self.name,
                args=args,
                server=self.server,
            )

    async def __httpstream_get(
        self,
        http: str,
        promptname: str,
        args: dict,
        oauth_client: OAuthClient | None,
        headers: dict[str, str] = None,
    ):
        oauth: OAuthClientProvider = (
            oauth_client.oauth if oauth_client is not None else None
        )
        async with streamablehttp_client(url=http, auth=oauth, headers=headers) as (
            read_stream,
            write_stream,
            _,
        ):
            async with ClientSession(read_stream, write_stream) as session:
                await session.initialize()

                # get a prompt
                prompt_result = await session.get_prompt(
                    name=promptname,
                    arguments=args,
                )
                return prompt_result.messages

    async def __stdio_get(
        self,
        promptname: str,
        args: dict,
        server: dict,
    ):
        server_params: StdioServerParameters = get_stdio_session_params(server)
        async with stdio_client(server_params) as (read, write):
            async with ClientSession(read, write) as session:
                await session.initialize()

                # get a prompt
                prompt_result = await session.get_prompt(
                    name=promptname,
                    arguments=args,
                )
                return prompt_result.messages



import uuid


class MessagesSet:
    def __init__(self, query: str):
        self.root_query: str = query
        self.id: str = str(uuid.uuid4())

        self.history_messages: list[dict] = []
        self.flow_messages: list[dict] = [{"query": query}]

    def append_message(self, message: dict):
        self.history_messages.append(message)
        self.flow_messages.append({"llm_message": message})

    def selected_prompts(self, prompts: list[dict]):
        self.flow_messages.append({"selected_prompts": prompts})

    def selected_service(self, service: dict):
        self.flow_messages.append({"selected_service": service})

    def selected_querys(self, querys: list[str]):
        self.flow_messages.append({"selected_querys": querys})

    @property
    def info(self) -> dict:
        return {
            "id": self.id,
            "history": self.history_messages,
            "flow": self.flow_messages,
        }


from collections.abc import Generator
from ..app.chat.features.step import Step


def print2terminal(chat_response: Generator[Step, None, None]):
    index = 1
    for step in chat_response:
        if step.type == "response":
            if step.json["first_chunk"]:
                print(f"<< {step.response}", end="")
            else:
                print(f"{step.response}", end="")
        if step.type == "query":
            print(f">> {step.query}")
            index = 1
        if step.type == "data":
            print(f'   {str(step).replace("**", "").replace("- ", "• ")}')
        if step.type == "step":
            print(f"  {index}. {step.step}")
            index += 1


def step2terminal(step: dict, index: int) -> int:
    if step["type"] == "response":
        if step["first_chunk"]:
            print(f"<< {step['response']}", end="")
        else:
            print(f"{step['response']}", end="")
    if step["type"] == "query":
        print(f">> {step['query']}")
        _index = 1
    if step["type"] == "data":
        print(f'   {str(step).replace("**", "").replace("- ", "• ")}')
    if step["type"] == "step":
        print(f"  {index}. {step['step']}")
        _index = index + 1

    return index



from .fastchat.api.api import FastApp, FastAPI

fastapp: FastApp = FastApp(
    extra_reponse_system_prompts=[
        "Eres un NPC de un vendedor ambulante para un juego RPG. Debes comportarte como tal y dar tus respuestas acorde a tu personaje. Te dedicas a la venta de armamento medieval, como espadas, armaduras, escudos y otros. Dirigete  quien te hable como si ese usuario fuera un aventurero en un mundo medieval de fantasia."
    ]
)
app: FastAPI = fastapp.app



import os
import json
import aiohttp
import requests
from urllib.parse import urljoin
from ..chat.message import MessagesSet
from ...config.logger import logger

ROOT_PATH = "http://127.0.0.1:6543/fastchatdb/"
SAVE_HISTORY = "history/save"
SAVE_MESSAGE = "message/save"
LOAD_HISTORY = "history/load"


class ClientDB:
    def __init__(
        self,
        config_file: str = "fastchat.config.json",
    ):
        self.base_body: dict = {}
        self.base_query: dict = {}

        self.is_none: bool = True
        self.headers: dict | None = None
        self.__load_config(config_file=config_file)

    def __load_config(self, config_file: str):
        """
        Load the configuration from the specified JSON file.
        If the file does not exist, use default values.
        """
        root_path: str = ROOT_PATH
        save_history: str = SAVE_HISTORY
        load_history: str = LOAD_HISTORY
        save_message: str = SAVE_MESSAGE

        if os.path.exists(config_file):
            with open(config_file, "r") as file:
                config: dict = json.load(file)

            db_connection: dict = config.get("db_conection")
            if db_connection is not None:
                self.is_none = False

                self.headers = db_connection.get("headers", {})
                self.base_body = db_connection.get("base_body", {})
                self.base_query = db_connection.get("base_query", {})

                root_path = db_connection.get("root_path") or ROOT_PATH

                endpoints: dict = db_connection.get("endpoints")
                if endpoints is not None:

                    save_history_endpoint: dict = endpoints.get("save_history")
                    if save_history_endpoint is not None:
                        save_history = save_history_endpoint.get("path") or save_history

                    load_history_endpoint: dict = endpoints.get("load_history")
                    if load_history_endpoint is not None:
                        load_history = load_history_endpoint.get("path") or load_history

                    save_message_endpoint: dict = endpoints.get("save_message")
                    if save_message_endpoint is not None:
                        save_message = save_message_endpoint.get("path") or save_message

        self.set_database_endpoints_path(
            root_path=root_path,
            save_history=save_history,
            load_history=load_history,
            save_message=save_message,
        )

    def set_database_endpoints_path(
        self,
        root_path: str = ROOT_PATH,
        save_history: str = SAVE_HISTORY,
        load_history: str = LOAD_HISTORY,
        save_message: str = SAVE_MESSAGE,
    ):
        if not root_path.endswith("/"):
            root_path += "/"

        save_history = (
            save_history[1:] if save_history.startswith("/") else save_history
        )
        load_history = (
            load_history[1:] if load_history.startswith("/") else load_history
        )
        save_message = (
            save_message[1:] if save_message.startswith("/") else save_message
        )

        self.save_history_path: str = urljoin(root_path, save_history)
        self.load_history_path: str = urljoin(root_path, load_history)
        self.save_messsage_path: str = urljoin(root_path, save_message)

    async def _post(self, url: str, json_data: dict) -> dict:
        """
        Asynchronously send a POST request to the given URL with the provided JSON data.
        Returns the response as a dictionary.
        """
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=json_data, headers=self.headers) as resp:
                return await resp.json()

    async def _get(self, url: str, params: dict) -> dict:
        """
        Asynchronously send a GET request to the given URL with the provided query parameters.
        Returns the response as a dictionary.
        """
        async with aiohttp.ClientSession() as session:
            async with session.get(url, params=params, headers=self.headers) as resp:
                return await resp.json()

    def load_history(self, chat_id: str) -> list:
        """
        Load the chat history from the database via the configured endpoint.
        Returns the history as a list if successful, otherwise an empty list.
        """
        if self.is_none:
            return []

        params = self.base_query.copy()
        params.update({"chat_id": chat_id})
        try:
            response = requests.get(
                self.load_history_path, params, headers=self.headers
            )
            if response.get("status") == "success":
                logger.info(f"History for chat_id {chat_id} loaded successfully")
                return response.get("history", [])
        except Exception as e:
            logger.warning(f"Error loading history for chat_id {chat_id}: {e}")
            return []

        return []

    async def save_history(self, chat_id: str, history: list[dict]) -> bool:
        """
        Asynchronously save the chat history to the database via the configured endpoint.
        Returns True if the operation was successful.
        """
        if self.is_none:
            return False

        data = self.base_body.copy()
        data.update({"chat_id": chat_id, "history": history})
        try:
            response = await self._post(self.save_history_path, data)
            logger.info(f"History for chat_id {chat_id} saved successfully")
            return response.get("status") == "success"
        except Exception as e:
            logger.warning(f"Error saving history for chat_id {chat_id}: {e}")
            return False

    async def save_message(
        self,
        chat_id: str,
        message_id: str,
        message: MessagesSet,
        user_id: str = "public",
    ) -> bool:
        """
        Asynchronously save a single message to the database via the configured endpoint.
        Returns True if the operation was successful.
        """
        if self.is_none:
            return False

        data = self.base_body.copy()
        data.update(
            {
                "chat_id": chat_id,
                "message_id": message_id,
                "user_id": user_id,
                "message": message.info,
            }
        )
        try:
            response = await self._post(self.save_messsage_path, data)
            if response.get("status") == "success":
                logger.info(f"Message {message_id} saved to database successfully")
            else:
                logger.warning(
                    f"Error saving message {message_id}: {response.get('message')}"
                )

            return response.get("status") == "success"
        except Exception as e:
            logger.warning(f"Error saving message {message_id}: {e}")

            return False  # Error



import json
import uuid
from typing import AsyncGenerator
from mcp.types import PromptMessage
from .message import MessagesSet
from ..client_db.client_db import ClientDB
from .features.llm_provider import LLMProvider
from .features.step import Step, StepMessage, DataStep, ResponseStep, QueryStep
from ..services.llm import LLM
from ..mcp_manager.client import ClientManagerMCP
from ..services.llm.models.openai_service.gpt import GPT
from ...config.llm_config import ConfigGPT, ConfigLLM


class Fastchat:
    """
    Chat class for managing conversational interactions with an LLM (Large Language Model).
    This class orchestrates the process of handling user queries, selecting prompts and services,
    and generating responses using a language model. It supports context management, prompt selection,
    service invocation, and streaming responses in a step-wise manner.
    Args:
        llm_provider (LLMProvider): The provider of language model to use. Default is `LLMProvider.OPENAI`.
        model (str): The model name to use for the LLM. Defaults to ConfigGPT.DEFAULT_MODEL_NAME.
        len_context (int): Maximum length of the conversation history to maintain. Defaults to 10.
        history (list): Initial chat history. Defaults to an empty list.
        id (str | None): Optional identifier for the chat session.
        user_id (str): Optional identifier for the chat session.
    
    ## Methods
        __call__(query: str) -> Generator[Step]:
            Processes a user query through multiple steps, including query analysis,
            prompt selection, service selection, and response generation. Yields Step objects
            representing each stage of the process.
        proccess_query(query: str) -> Generator[Step]:
            Handles the detailed processing of a single query, including updating chat history,
            selecting prompts and services, and generating the final response. Yields Step objects
            for each sub-step in the process.
    """

    def __init__(
        self,
        id: str | None = None,
        model=ConfigGPT.DEFAULT_MODEL_NAME,
        llm_provider: LLMProvider = ConfigLLM.DEFAULT_PROVIDER,
        extra_reponse_system_prompts: list[str] = [],
        extra_selection_system_prompts: list[str] = [],
        aditional_servers: dict = {},
        len_context: int = ConfigLLM.DEFAULT_HISTORY_LEN,
        history: list = [],
        user_id: str = "public",
    ):
        """
        Initialize a Chat instance.
        Args:
            id (str | None, optional): Optional chat session identifier.
            model (str, optional): The model name to use. Defaults to ConfigGPT.DEFAULT_MODEL_NAME.
            llm_provider (LLMProvider, optional): The language model provider. Defaults to LLMProvider.OPENAI.
            extra_reponse_system_prompts (list[str]): Additional system prompts for responses. Defaults to an empty list.
            extra_selection_system_prompts (list[str]): Additional system prompts for MCP services selection. Defaults to an empty list.
            aditional_servers (dic, optional): dictionary of servers with the format: `{server_1:{...server config...}, "server_2":{...}}`
            len_context (int, optional): Maximum number of messages to keep in context. Defaults to 10.
            history (list, optional): Initial chat history. Defaults to empty list.
        """

        self.clientdb: ClientDB = ClientDB()
        self.user_id: str = user_id

        self.id = id if id is not None else str(uuid.uuid4())
        loaded_history = (
            self.clientdb.load_history(chat_id=self.id) if id is not None else []
        )

        self.llm: LLM = GPT(
            max_history_len=len_context,
            model=model,
            chat_history=history + loaded_history,
        )
        self.client_manager_mcp: ClientManagerMCP | None = None  # MCP client manager for handling server connections
        self.extra_reponse_system_prompts: list[dict[str, str]] = [
            {
                "role": "system",
                "content": message,
            }
            for message in extra_reponse_system_prompts
        ]
        self.extra_selection_system_prompts: list[dict[str, str]] = [
            {
                "role": "system",
                "content": message,
            }
            for message in extra_selection_system_prompts
        ]
        self.aditional_servers: dict = aditional_servers

        self.current_messages_set: MessagesSet | None = None

    async def close(self) -> None:
        """
        Properly cleanup and close the Fastchat instance.
        
        This method ensures all active connections and resources are properly closed
        to prevent memory leaks and hanging connections. It should be called when 
        the chat session is no longer needed.
        
        Changes made for cleanup implementation:
        - Added proper MCP client manager cleanup
        - Clear all reference collections to help garbage collection
        - Set client manager to None to prevent further usage
        """
        # Close MCP client manager and all its connections
        # This will close all HTTP/WebSocket connections and stdio processes
        if self.client_manager_mcp is not None:
            await self.client_manager_mcp.close()  # Closes all MCP server connections
            self.client_manager_mcp = None  # Prevent further usage

        # Close LLM client (if tiene método close)
        if hasattr(self.llm, 'close') and callable(getattr(self.llm, 'close')):
            await self.llm.close()
        
        # Clear references to help garbage collection
        # These collections might hold references to large objects
        self.current_messages_set = None
        self.extra_reponse_system_prompts.clear()
        self.extra_selection_system_prompts.clear()
        self.aditional_servers.clear()

    async def __aenter__(self):
        """
        Context manager entry point.
        
        This method is part of the async context manager protocol implementation.
        It automatically initializes the Fastchat instance when entering a 
        'async with' block, ensuring proper setup.
        
        Returns:
            self: The initialized Fastchat instance
        """
        await self.initialize()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Context manager exit point.
        
        This method is part of the async context manager protocol implementation.
        It automatically calls close() when exiting an 'async with' block,
        ensuring proper cleanup even if an exception occurs.
        
        Args:
            exc_type: Exception type (if any)
            exc_val: Exception value (if any) 
            exc_tb: Exception traceback (if any)
        """
        await self.close()

    
    async def initialize(self, print_logo: bool = True) -> None:
        await self.__set_client_manager_mcp(print_logo=print_logo)

    async def __set_client_manager_mcp(self, print_logo: bool) -> None:
        if self.client_manager_mcp is None:
            self.client_manager_mcp = ClientManagerMCP(
                aditional_servers=self.aditional_servers,
                print_logo=print_logo,
            )
            await self.client_manager_mcp.initialize()
            self.llm.set_client_manager_mcp(self.client_manager_mcp)

    ####  FLOW   FLOW   FLOW ####
    async def __call__(self, query: str) -> AsyncGenerator[Step]:
        """
        Processes a user query through the chat pipeline.
        This method analyzes the query, preprocesses it, selects relevant prompts and services,
        and generates a response. The process is broken down into steps, each represented by a Step object,
        which are yielded sequentially to allow for streaming or step-wise processing.
        Args:
            query (str): The user's input query.
        Yields:
            Step: An object representing each stage of the query processing pipeline.
        """

        yield Step(step_type=StepMessage.ANALYZE_QUERY)
        self.current_messages_set = MessagesSet(query=query)

        processed_query: dict = self.llm.preprocess_query(query=query)
        querys: list[str] = processed_query["querys"]
        self.llm.current_language = processed_query["language"]

        self.current_messages_set.selected_querys(querys=querys)

        yield DataStep(data={"querys": querys})

        for query in querys:
            async for step in self.proccess_query(query):
                yield step

        # Save To Database Here
        messsages2save: MessagesSet = self.current_messages_set
        await self.clientdb.save_message(
            chat_id=self.id,
            message_id=messsages2save.id,
            user_id=self.user_id,
            message=messsages2save,
        )
        ###########################

        # self.current_messages_set = None

    async def proccess_query(self, query: str) -> AsyncGenerator[Step]:
        """
        Handles the detailed processing of a single query.
        This method updates the chat history, selects prompts and services based on the query,
        invokes the appropriate service or tool, and generates the final response. Each sub-step
        is yielded as a Step object for granular control and streaming.
        Args:
            query (str): The user's input query.
        Yields:
            Step: An object representing each sub-step of the query processing, including prompt selection,
                  service selection, and response generation.
        """
        client_manager_mcp: ClientManagerMCP = self.client_manager_mcp
        self.llm.append_chat_history()
        yield QueryStep(query)

        self.current_messages_set.append_message({"role": "user", "content": query})

        # region ########### GET PROMTPS ###########
        yield Step(step_type=StepMessage.SELECT_PROMPTS)
        prompts = json.loads(
            self.llm.select_prompts(query, self.extra_selection_system_prompts)
        )["prompt_services"]
        self.current_messages_set.selected_prompts(prompts=prompts)

        if len(prompts) == 0:
            yield DataStep(data={"prompts": None})

        for index, prompt in enumerate(prompts):
            yield DataStep(data={f"prompt {index+1}": prompt["prompt_service"]})

        extra_messages = [
            await client_manager_mcp.prompts[prompt["prompt_service"]](prompt["args"])
            for prompt in prompts
        ]
        extra_messages = [
            prompt_message2dict(message)
            for prompt_message in extra_messages
            for message in prompt_message
        ]
        # endregion ###################################################

        # region ########### SELECT SERVICE AND ARGS ############
        yield Step(step_type=StepMessage.SELECT_SERVICE)
        # Cargar los servicios utiles
        service = json.loads(
            self.llm.select_service(
                query,
                extra_messages=extra_messages + self.extra_selection_system_prompts,
            )
        )
        args = service["args"]
        service = service["service"]
        self.current_messages_set.selected_service(
            service={"service": service, "args": args}
        )
        # endregion ###################################################

        # region ########### RESPONSE ############
        if len(service) == 0:
            yield DataStep(data={"service": None})
            first_chunk = True
            stream = self.llm.simple_query(
                query,
                use_services_contex=True,
                extra_messages=extra_messages + self.extra_reponse_system_prompts,
            )
            response = ""
            for chunk in stream:
                yield ResponseStep(response=chunk, data=None, first_chunk=first_chunk)
                response += chunk if chunk is not None else ""
                first_chunk = False

            self.current_messages_set.append_message(
                {"role": "assistant", "content": response}
            )
            yield ResponseStep(response="\n\n", data=None)

        else:
            yield DataStep(data={"service": service})
            yield DataStep(data={"args": args})
            service = (
                client_manager_mcp.resources[service]
                if (client_manager_mcp.service_type(service) == "resource")
                else (
                    client_manager_mcp.tools[service]
                    if (client_manager_mcp.service_type(service) == "tool")
                    else None
                )
            )

            data = await service(args)
            data = data[0].text
            first_chunk = True
            for chunk in self.llm.final_response(
                query,
                data,
                extra_messages=self.extra_reponse_system_prompts,
            ):
                yield ResponseStep(response=chunk, data=None, first_chunk=first_chunk)
                first_chunk = False
            yield ResponseStep(response="\n\n", data=data)

        # endregion ###################################################


def prompt_message2dict(prompt_message: PromptMessage):
    data: dict = json.loads(prompt_message.content.text)
    return {"role": prompt_message.role, "content": data["content"]["text"]}


""" FLOW
see https://github.com/rb58853/fastchat-mcp/tree/main/doc/FLOW.md
"""



import asyncio
from .servers import Servers
from .sessions import httpstrem, stdio
from .services import Tool, Resource, Prompt
from ...config.logger import logger, CustomFormatter, LoggerFeatures


class ClientManagerMCP:
    """
    ClientManagerMCP is responsible for managing and interacting with tools, resources, and prompts
    provided by multiple MCP (Multi-Component Platform) servers. It handles the initialization,
    refreshing, and retrieval of these components, as well as facilitating their invocation.
    
    CLEANUP IMPLEMENTATION CHANGES:
    - Added close() method to properly shutdown all MCP server connections
    - Added context manager support (__aenter__, __aexit__) for automatic cleanup
    - Implemented __close_all_sessions() to handle both HTTP and stdio session cleanup
    - Clear all cached data (tools, resources, prompts) on close to prevent memory leaks
    
    Attributes:
        app_name (str): The name of the application using the manager.
        tools (dict[str, Tool]): Dictionary of available tools, keyed by server and tool name.
        resources (dict[str, Resource]): Dictionary of available resources, keyed by server and resource name.
        prompts (dict[str, Prompt]): Dictionary of available prompts, keyed by server and prompt name.
    ## Methods
        close(): NEW - Properly close all MCP server connections and cleanup resources
        call_tool(name: str, args: dict) -> str | None:
            Calls a tool by name with the provided arguments.
        read_resource(name: str, args: dict) -> str | None:
            Reads a resource by name with the provided arguments.
        get_prompt(name: str, args: dict) -> dict:
            Retrieves a prompt by name with the provided arguments.
        refresh_data():
            Initializes or refreshes the lists of tools, resources, and prompts from all registered MCP servers.
        service_type(service_key: str) -> str:
            Determines whether a given service key corresponds to a tool or a resource.
        get_services() -> list[dict[str, any]]:
            Returns a list of dictionaries representing the available services.
        get_prompts() -> list[dict[str, any]]:
            Returns a list of dictionaries representing the available prompts.
    ## Private Methods
        __get_session(server: dict) -> dict:
            Establishes a session with a given server and retrieves its available tools, resources, and prompts.
    ## Usage
        Instantiate ClientManagerMCP to manage and interact with MCP server components in a unified way.
    """

    INSTANCE = None

    def __init__(
        self,
        aditional_servers: dict = {},
        print_logo: bool = True,
    ):
        self.tools: dict[str, Tool] | None = {}
        self.resources: dict[str, Resource] | None = {}
        self.prompts: dict[str, Prompt] | None = {}

        self.__services: list[dict] = []
        """List of services as strings to be passed to the LLM"""
        self.__prompts_context: list[dict] = []
        """List of prompts as strings to be passed to the LLM"""
        self.__print_logo: bool = print_logo

        self.aditional_servers: dict = aditional_servers

    async def initialize(self) -> None:
        await self.refresh_data()

    async def refresh_data(self) -> None:
        """Initialize or refresh the lists of tools, resources, and prompts from all registered MCP servers."""
        # This method should be implemented to populate tools, resources, and prompts
        pass
        ClientManagerMCP.INSTANCE = self

    # region SINGLETON
    # async def initialize(self) -> None:
    #     if ClientManagerMCP.INSTANCE is None:
    #         await self.refresh_data()
    #         ClientManagerMCP.INSTANCE = self
    #     else:
    #         self.tools = ClientManagerMCP.INSTANCE.tools
    #         self.resources = ClientManagerMCP.INSTANCE.resources
    #         self.prompts = ClientManagerMCP.INSTANCE.prompts
    # endregion

    
    async def close(self) -> None:
        """
        Properly close all MCP server connections and cleanup resources.
        
        NEW METHOD ADDED FOR CLEANUP IMPLEMENTATION:
        This method ensures all active MCP sessions are properly terminated
        and all cached data is cleared to prevent memory leaks.
        
        What it does:
        1. Closes all HTTP stream sessions (WebSocket/HTTP connections)
        2. Closes all stdio sessions (subprocess connections)  
        3. Clears all cached tools, resources, and prompts
        4. Clears internal service and prompt context lists
        5. Resets singleton instance to None
        """
        # Close all active sessions with MCP servers
        await self.__close_all_sessions()
        
        # Clear all cached data to prevent memory leaks
        self.tools.clear() if self.tools else None
        self.resources.clear() if self.resources else None
        self.prompts.clear() if self.prompts else None
        
        # Clear internal service caches
        self.__services.clear()
        self.__prompts_context.clear()
        
        # Clear singleton instance reference
        ClientManagerMCP.INSTANCE = None
        
        logger.info("ClientManagerMCP closed successfully")

    async def __close_all_sessions(self) -> None:
        """
        Close all active sessions with MCP servers.
        
        NEW METHOD ADDED FOR CLEANUP IMPLEMENTATION:
        This method iterates through all configured MCP servers and calls
        the appropriate close_session function based on the protocol type.
        
        Handles two types of MCP connections:
        1. httpstream - HTTP/WebSocket connections 
        2. stdio - Subprocess connections
        """
        mcp_servers: dict[str, dict] = Servers(aditional_servers=self.aditional_servers).mcp_servers or {}
        
        for server_key in mcp_servers.keys():
            server = {"key": server_key} | mcp_servers[server_key]
            
            try:
                if server["protocol"] == "httpstream":
                    # Close HTTP/WebSocket session
                    await httpstrem.close_session(server)
                elif server["protocol"] == "stdio":
                    # Close subprocess session
                    await stdio.close_session(server)
                    
                logger.info(
                    f"Closed connection with server {CustomFormatter.green}{server_key}{CustomFormatter.reset}"
                )
            except Exception as e:
                logger.error(
                    f"Error closing connection with server {CustomFormatter.bold_red}{server_key}{CustomFormatter.reset}: {e}"
                )

    async def __aenter__(self):
        """
        Context manager entry point.
        
        NEW METHOD ADDED FOR CLEANUP IMPLEMENTATION:
        Enables async context manager support for automatic resource management.
        When entering an 'async with' block, this method initializes the 
        ClientManagerMCP instance.
        
        Returns:
            self: The initialized ClientManagerMCP instance
        """
        await self.initialize()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Context manager exit point.
        
        NEW METHOD ADDED FOR CLEANUP IMPLEMENTATION:
        Enables async context manager support for automatic resource management.
        When exiting an 'async with' block, this method automatically calls
        close() to ensure proper cleanup, even if an exception occurs.
        
        Args:
            exc_type: Exception type (if any)
            exc_val: Exception value (if any)
            exc_tb: Exception traceback (if any)
        """
        await self.close()

    async def call_tool(self, name: str, args: dict) -> str | None:
        return await self.tools[name](args)

    async def read_resource(self, name: str, args: dict) -> str | None:
        return await self.resources[name](args)

    async def get_prompt(self, name: str, args: dict) -> dict:
        return await self.prompts[name](args)

    async def refresh_data(self):
        """
        ### Refresh Datas
        - Initializes or refreshes the lists of tools, resources, and prompts provided by each mcp server,
        organizing them into dictionaries.
        - For each registered MCP servers, retrieves their available tools, resources,
        and prompts via a session, and stores them in the corresponding self dictionaries.
        - If a session cannot be established with a server, that server is skipped.
        """
        if self.__print_logo:
            print(LoggerFeatures.LOGO)

        self.tools, self.resources, self.prompts = ({}, {}, {})
        mcp_servers: dict[str, dict] = Servers(aditional_servers=self.aditional_servers).mcp_servers

        for server_key in mcp_servers.keys():
            server = {"key": server_key} | mcp_servers[server_key]
            session = await self.__get_session(server)
            if session is None:
                continue

            for tool in session["tools"]:
                self.tools[f"{server_key}_{tool.name}"] = Tool(data=tool, server=server)
            for resource in session["resources"]:
                self.resources[f"{server_key}_{resource.name}"] = Resource(
                    data=resource, server=server
                )

            for prompt in session["prompts"]:
                self.prompts[f"{server_key}_{prompt.name}"] = Prompt(
                    data=prompt, server=server
                )

    async def __get_session(self, server: dict) -> dict:
        try:
            session: dict = {}

            if server["protocol"] == "httpstream":
                session = await httpstrem.async_get_session(
                    server["httpstream-url"],
                    server["oauth_client"],
                    headers=server.get("headers", None),
                )
            elif server["protocol"] == "stdio":
                session = await stdio.async_get_session(server=server)
            else:
                logger.error(
                    f"Unsupported protocol type {CustomFormatter.bold_red}{server['protocol']}{CustomFormatter.reset} for server {server['key']}"
                )
                return None

            logger.info(
                f"Establish connection with server {CustomFormatter.green}{server['key']}{CustomFormatter.reset} successfully."
            )
            return session
        except Exception as e:
            logger.error(
                f"Failed to establish connection with server {CustomFormatter.bold_red}{server['key']}{CustomFormatter.reset}. Cause: {e}"
            )
            return None

    def service_type(self, service_key: str) -> str:
        """
        Determine the type of service associated with the given service key.
        Args:
            service_key (str): The key identifying the service.
        Returns:
            str: "tool" if the service key corresponds to a tool, "resource" if it corresponds to a resource.
        """

        if service_key in self.tools.keys():
            return "tool"
        if service_key in self.resources.keys():
            return "resource"

    def get_services(self) -> list[dict[str, any]]:
        """
        Returns a list of dictionaries representing the available services. Creates
        a dictionary for each service with the service name as the key and its string
        representation as the value.
        Returns:
            list[dict]: A list of dictionaries, each containing a service name
            and its corresponding string representation.
        """

        if len(self.__services) != len(self.tools) + len(self.resources):
            services = self.tools | self.resources
            self.__services = [
                {service: str(services[service])} for service in services.keys()
            ]

        return self.__services

    def get_prompts(self) -> list[dict[str, any]]:
        """
        Returns a list of prompt dictionaries, ensuring each prompt is represented as a dictionary
        with string values.
        Returns:
            list[dict]: A list of dictionaries containing the current prompts.
        """

        if len(self.__prompts_context) != len(self.prompts):
            self.__prompts_context = [
                {key: str(self.prompts[key])} for key in self.prompts.keys()
            ]
        return self.__prompts_context
